###############################################################################
## DESCRIPTION
###############################################################################

# Grand Canonical Monte Carlo using the Wang-Landau method to get a 
# histogram flat in the number of particles, N

#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.

###############################################################################
## IMPORT MODULES
###############################################################################

#import python modules
import numpy as np
import sys
import logging
import argparse
import atomwrite

logging.basicConfig(stream=sys.stdout, format='%(levelname)s:%(message)s',
										level=logging.INFO)
#logging.basicConfig()

###############################################################################
## ARGUMENT LIST
###############################################################################
#Parse command-line options
parser = argparse.ArgumentParser(description='Grand Canonical Monte Carlo ' +
				 'using the Wang-Landau method to get a histogram flat in N')
parser.add_argument('--Temp', '-t', type=float, required=True,
                    help='dimensionless temperature')
parser.add_argument('--Rhomax', '-r', type=float, required=True,
                    help='maximum particle density')
parser.add_argument('--ForceField', '-f', required=True,
                    help='select a force field for pair interactions: LJ, SQ, SSQ')
parser.add_argument('--Lambda', '-l', type=float,
                    help='width of square well potential')
args = parser.parse_args(sys.argv[1:])
logging.info('Temp=' + str(args.Temp))
logging.info('rhomax=' + str(args.Rhomax))
logging.info('ForceField=' + args.ForceField)
if args.ForceField == 'LJ':
	import LJlib as FFlib
	Cut2 = (2.5) ** 2
else:
	if args.ForceField == 'SQ':
		import SQlib as FFlib
	elif args.ForceField == 'SSQ':
		import SSQlib as FFlib
	logging.info('lambda=' + str(args.Lambda))
	Cut2 = (args.Lambda) ** 2

###############################################################################
## DEFINE FUNCTIONS
###############################################################################

def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def ParticleMove(Pos, U, N):
	"""Displace a single particle Monte Carlo style using Metropolis criteria
Input:
		Pos: (Nmax,3) array of positions
		U: potential energy
		N: number of particles
Output:
		Pos: (Nmax,3) array of positions
		U: potential energy
		Accept: Boolean value whether move was accepted
"""
	#trial move
	MoveAtom = np.random.randint(N)
	OldPos = Pos[MoveAtom,:].copy()
	OldEnergy = FFlib.calcparticleenergy(MoveAtom, Pos, N, L, Cut2)
	Pos[MoveAtom,:] = OldPos + np.random.uniform(-dx, dx, 3)
	NewEnergy = FFlib.calcparticleenergy(MoveAtom, Pos, N, L, Cut2)
	dU = NewEnergy - OldEnergy

	if args.ForceField == 'SQ' and NewEnergy > 0:
		#reject core overlap
		Accept = False
	else:
		#use Metropolis criteria
		#accept the move unless it's uphill, then accept proportionately
		Accept = True
		if B * dU > 0:
			if np.random.rand() > np.exp(- B * dU):
				Accept = False

	#reset if move is rejected
	if not Accept:
		Pos[MoveAtom,:] = OldPos

	#update energy if move is accepted
	if Accept:
		U += dU
	
	logging.debug("B=" + str(B))
	logging.debug("MoveAtom=" + str(MoveAtom))
	logging.debug("OldEnergy=" + str(OldEnergy))
	logging.debug("NewEnergy=" + str(NewEnergy))
	logging.debug("Accept=" + str(Accept) + "\n")
	return Pos, U, Accept

def ParticleInsertion(Pos, U, N, W):
	"""Displace a single particle Monte Carlo style using Metropolis criteria
Input:
    Pos: (Nmax,3) array of positions
		U: potential energy
		N: number of particles
		W: (Nmax+1) array of sampled weights
Output:
		Pos: (N,3) array of positions
		U: potential energy
		N: number of particles
		Accept: Boolean value whether move was accepted
"""
	#trial location
	InsertAtom = N
	Pos[InsertAtom,:] = np.random.uniform(-L/2, L/2, 3)
	NewEnergy = FFlib.calcparticleenergy(InsertAtom, Pos, N+1, L, Cut2)
	dU = NewEnergy
	dW = W[N] - W[N-1]

	if args.ForceField == 'SQ' and NewEnergy > 0:
		#reject core overlap
		Accept = False
	else:
		#use Metropolis criteria
		#accept the move unless it's uphill, then accept proportionately
		Accept = True
		if B * dU > np.log(1. * V / (N+1)) + B * mu + dW:
			if np.random.rand() > 1. * V / (N+1) * np.exp(- B * dU + B * mu + dW):
				Accept = False

	#update particle number, energy if move is accepted
	if Accept:
		N += 1
		U += dU 

	logging.debug("InsertAtom=" + str(InsertAtom))
	logging.debug("NewEnergy=" + str(NewEnergy))
	logging.debug("Accept=" + str(Accept) + "\n")
	return Pos, U, N, Accept

def ParticleDeletion(Pos, U, N, W):
	"""Displace a single particle Monte Carlo style using Metropolis criteria
Input:
    Pos: (Nmax,3) array of positions
		U: potential energy
		N: number of particles
		W: (Nmax+1) array of sampled weights
Output:
		Pos: (Nmax,3) array of positions
		U: potential energy
		N: number of particles
		Accept: Boolean value whether move was accepted
"""
	#trial deletion
	RemoveAtom = np.random.randint(N)
	OldEnergy = FFlib.calcparticleenergy(RemoveAtom, Pos, N, L, Cut2)
	dU = - OldEnergy
	dW = W[N-2] - W[N-1]

	#use Metropolis criteria
	#accept the move unless it's uphill, then accept proportionately
	Accept = True
	if B * (dU + mu) > np.log(1. * N / V) + dW:
		if np.random.rand() > 1. * N / V * np.exp(- B * dU - B * mu + dW):
			Accept = False

	#update particle number, energy if move is accepted
	if Accept:
		N -= 1
		U += dU
		Pos[RemoveAtom:Nmax-1,:] = Pos[RemoveAtom+1:Nmax,:]
		Pos[Nmax-1,:] = np.zeros(3, float)

	logging.debug("RemoveAtom=" + str(RemoveAtom))
	logging.debug("OldEnergy=" + str(OldEnergy))
	logging.debug("Accept=" + str(Accept) + "\n")
	return Pos, U, N, Accept

def Reweight(W, mu):
	"""
Input:
		W: biasing function
		mu: chemical potential 
Output:
		p: probability distribution
"""
	p = [np.exp(-w + B * mu) for w in W]
	p /= sum(p)
	return p.tolist()

###############################################################################
## SET VARIABLES
###############################################################################

#ENSEMBLE CONDITIONS
mu = 0.
V = 200; L = V ** (1./3.)
B = 1. / args.Temp
#mu = chemical potential, g/epsilon
#V = volume; L = box dimension
#B = inverse temperature, epsilon/kT

#RANGE OF FLUCTUATIONS
Nmin = 1
Nmax = int(args.Rhomax * V)
logging.info('Nmax=' + str(Nmax))
#Nmin = minimum number of particles in the box
#Nmax = maximum number of particles in the box

#INITIAL CONDITIONS
N = 1
G = 1.
#N = number of total atoms in the system
#G = modification factor, incremental weight for states visited

#PARAMETERS
dx = 0.10
mEquil = 0
GTol = 1.e-6
OutputFreq = 1e6
FullEnergyFreq = 5e4
Rmove = 0.5
Rinsert = (1. - Rmove) / 2. + Rmove
# dx = maximum step size
# mEquil = number of equilibration moves, particle displacements only
# OutputFreq = number of moves between reporting energies
# FullEnergyFreq = number of moves between full energy updates
# Rmove = portion of all moves that are displacement moves
# Rinsert = 

#INITIALIZE ARRAYS
Pos = np.zeros((Nmax, 3), float)
W = np.zeros(Nmax, float)
C = np.zeros(Nmax, int)
p = [0] * (Nmax)
#Pos = atomic positions
#W = weights, 
#C = counts,
#p = probability distribution of N

###############################################################################
## MODULE
###############################################################################
#np.random.seed(342324)

logging.info("#Initialization")
Pos[:N,:] = InitPositions(N, L)
U = FFlib.calcenergy(Pos, N, L, Cut2)
logging.debug("N,U=" + str(N) + "," + str(U) + "\n")

totAttempt = {'move':0, 'insert':0, 'delete':0}
totAccept = totAttempt.copy()
avgAccept = totAttempt.copy()

if mEquil > 0:
	logging.info("#Equilibration")
	logging.debug("%-6s %6s %6s %6s %6s" % ('#move', 'Part', 'Energy', 
					'mAcc', 'dx'))
	logging.debug("%-6d %6d %6f %6s %6.3f" % (0, N, U, '--', dx))
	for m in range(1,mEquil+1):
		logging.debug("Attempting move")
		Pos, U, Accept = ParticleMove(Pos, U, N)
		totAttempt['move'] += 1
		totAccept['move'] += Accept
		if m % FullEnergyFreq == 0:
			U = FFlib.calcenergy(Pos, N, L, Cut2)
			logging.debug('U=' + str(U) + '\n')
		if m % OutputFreq == 0:
			avgAccept['move'] = 1.0 * totAccept['move'] / totAttempt['move']
			totAccept['move'] = 0
			totAttempt['move'] = 0

			#adjust stepsize
			if avgAccept['move'] >= 0.7:
				dx *= 1.1
			elif avgAccept['move'] <= 0.3:
				dx *= 0.9

			logging.debug("%-6d %6d %6f %6.3f %6.3f" % (m/OutputFreq, N, U, 
							avgAccept['move'], dx))

m = 0
if G > GTol: 
	logging.info("#Production")
	logging.info("%-6s %6s %6s %6s %6s %6s %6s" % ('#move', 'Part', 'Energy', 
					'iAcc', 'dAcc', 'mAcc', 'dx'))
while G > GTol:
	m += 1

	#propose a move
	R = np.random.rand()
	if R < Rmove: 
		Move='move'
		logging.debug("Attempting move")
		Pos, U, Accept = ParticleMove(Pos, U, N)
	elif R >= Rmove and R < Rinsert: 
		Move='insert'
		logging.debug("Attempting insertion")
		if N < Nmax:
			Pos, U, N, Accept = ParticleInsertion(Pos, U, N, W)
		else:
			Accept = False
	elif R >= Rinsert: 
		Move='delete'
		logging.debug("Attempting deletion")
		if N > Nmin:
			Pos, U, N, Accept = ParticleDeletion(Pos, U, N, W)
		else:
			Accept = False

	#update counters
	totAttempt[Move] += 1
	totAccept[Move] += Accept
	C[N-1] += 1
	W[N-1] -= G
	logging.debug("N,U=" + str(N) + "," + str(U) + "\n")

	#update energy
	if m % FullEnergyFreq == 0:
		U = FFlib.calcenergy(Pos, N, L, Cut2)
		logging.debug('U=' + str(U) + '\n')

	#output current system stats
	if m % OutputFreq == 0:
		for Move in totAttempt:
			if totAttempt[Move] != 0:
				avgAccept[Move] = 1.0 * totAccept[Move] / totAttempt[Move]
			else:
				avgAccept[Move] = 0.
			totAccept[Move] = 0
			totAttempt[Move] = 0

		#adjust stepsize
		if avgAccept['move'] >= 0.7:
			dx *= 1.1
		elif avgAccept['move'] <= 0.3:
			dx *= 0.9

		logging.info("%-6d %6d %6.3g %6.3f %6.3f %6.3f %6.3f" % (m/OutputFreq, N, 
						U, avgAccept['insert'], avgAccept['delete'], avgAccept['move'], dx))
		logging.debug('Counts = \n' + str(C))
		logging.debug('Bias   = \n' + str(W))
		logging.debug('Sampled bins = ' + str(len(C.nonzero()[0])) + '/' + 
						str(len(C)))
		logging.debug('Min: Count[' + str(np.where(C == C.min())[0][0]) + 
						'] = ' + str(C.min()))
		logging.debug('Avg: Count = ' + str(C.mean()))
		Occ = np.array([1]*N + [0]*(Nmax-N))
		Beta = np.zeros((Nmax), float)
		for n in range(N):
			Beta[n] = FFlib.calcparticleenergy(n, Pos, N, L, Cut2)
#		atomwrite.WritePdb('pdb/traj.pdb', Pos, Occ, Beta, L)

	#is histogram flat?
	if C.min() > 0.8 * C.mean():
		for Move in totAttempt:
			if totAttempt[Move] != 0:
				avgAccept[Move] = 1.0 * totAccept[Move] / totAttempt[Move]
			else:
				avgAccept[Move] = 0.
			totAccept[Move] = 0
			totAttempt[Move] = 0
		logging.info("%-6d %6d %6.3g %6.3f %6.3f %6.3f %6.3f" % (m, N, U,
						avgAccept['insert'], avgAccept['delete'], avgAccept['move'], dx))
		m = 0

		logging.info('Counts = \n' + str(C))
		C = np.zeros(Nmax, int)
		G /= 2.
		W = [w - W[0] for w in W]
		logging.info('Bias   = \n' + str(W) + '\n')
		logging.info('Resetting histogram. G=' + str(G))

	if m == 10:
		break

W = [w - W[0] for w in W]
logging.info('Final Bias = ' + str(W))

f = open('WL.dat', 'w')
f.write("%-6s %-6s\n" % ('#N', 'bA'))
for n in range(1,Nmax+1):
	f.write("%-6d %-6.3f\n" % (n, W[n-1]))
f.close()
